edition = "2023";

// LINT: LEGACY_NAMES

package protobuf_editions_test.edition2023;

message BadExtensionMessage {
  extensions 1 to max;
}

extend BadExtensionMessage {
  string BadExtensionName = 1;
}
