// Protocol Buffers - Google's data interchange format
// Copyright 2026 Google LLC.  All rights reserved.
//
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file or at
// https://developers.google.com/open-source/licenses/bsd

edition = "2023";

package third_party_protobuf_rust_test;

import "rust/test/extensions.proto";

option java_multiple_files = true;

extend TestExtensions {
  int32 i32_extension_separate_file = 1000;
}
