edition = "2023";

// LINT: LEGACY_NAMES

package protobuf_editions_test.edition2023;

message BadMethodMessage {
  string foo = 1;
}

service BadMethodService {
  // This is a method with a bad method name.
  rpc bad_method_name(BadMethodMessage) returns (BadMethodMessage) {}
}
